<?php
namespace taskman;
use Exception;

task('composer', function(array $args) {
  composer_run($args);
});

task('composer_reset', function()
{
  global $GAME_ROOT;

  echo "Going to reset composer package manager...\n";

  are_you_sure();

  ensure_rm("$GAME_ROOT/composer/vendor/");
  ensure_rm("$GAME_ROOT/composer/composer.last");
});

task('composer_vendor_push', function(array $args) {

  $usage = "Usage: ./gamectl composer_vendor_push <path/to/repo> <commit message> [^[^[^]]]\n";

  $dry_run = arg_opt($args, '--dry-run', false);
  arg_opt_check_no_trailing($args);

  if(count($args) < 2) 
    throw new Exception($usage);

  $force_new_tag = '';
  $up_mode = 1;
  if(count($args) > 2)
  {
    $tmp_arg = array_pop($args);

    if($tmp_arg === '^')
      $up_mode = 1;
    else if($tmp_arg === '^^')
      $up_mode = 2;
    else if($tmp_arg === '^^^')
      $up_mode = 3;
    else 
      $force_new_tag = $tmp_arg;
  }

  $relpath = array_shift($args);
  $commit_msg = array_shift($args);
  $repo = realpath("composer" .DIRECTORY_SEPARATOR. "vendor" .DIRECTORY_SEPARATOR. $relpath) ?: realpath($relpath);

  if(!$repo)
    throw new Exception("Failed to find a valid directory for arg '$relpath'\n");

  echo "Detected folder: '$repo'\n";

  if(!git_is_repo($repo, true))
    throw new Exception("'$repo' is not a valid Git repo!\n");

  echo "It's a valid Git repo\n";

  echo "======== Git status:\n";
  $status = [];
  git_do($repo, 'status', $status, false);
  echo implode("\n", $status) . "\n";
  echo "========\n";

  $last_remote_tag = get_git_last_remote_tag($repo);
  if($last_remote_tag === false)
    throw new Exception("Could not retrieve last remote tag");
  echo "The last remote Git tag: $last_remote_tag\n";

  if(empty($force_new_tag))
  {
    $last_remote_version = GitVersion::parse($last_remote_tag);
    $last_remote_version->bump($up_mode);
    $new_tag = $last_remote_version->encode();
  }
  else
    $new_tag = $force_new_tag;

  echo "Commit message: {$commit_msg}\n";
  echo "New version expected to be: {$new_tag}\n";

  if($dry_run === false)
    are_you_sure();

  $is_upm = file_exists("$repo/package.json");
  if($is_upm)
  {
    echo "This looks like an UPM package, let's update the package.json?\n";
    if(are_you_sure_ask())
    {
      if(!function_exists('taskman\upm_get_package_info'))
      {
        echo "Can't proceed, composer 'bit/taskman_upm' package is missing\n";
        exit(1);
      }

      $upm_info = upm_get_package_info($repo);
      if(!isset($upm_info['name']))
        throw new Exception("'name' is not present in package.json");
      if(!isset($upm_info['version']))
        throw new Exception("'version' is not present in package.json");
      $upm_version = $upm_info['version'];

      echo "Current package.json version is: {$upm_version}\n";
      $upm_new_version = ltrim($new_tag, 'v');
      echo "New package.json expected to be: {$upm_new_version}\n";

      if($dry_run === false)
      {
        are_you_sure();

        $upm_info['version'] = $upm_new_version;
        upm_set_package_info($repo, $upm_info);
      }
    }
  }

  if($dry_run !== false)
  {
    echo "Dry run is done.\n";
    return;
  }

  git_do($repo, 'add .');
  git_do($repo, "commit -am \"$commit_msg\"", $commit_output);
  $rev = git_rev_from_commit_message(trim($commit_output[0]));
  echo "Commited revision hash: $rev\n";
  if($rev == null || strlen($rev) < 6)
  {
    echo "Failed to detect a new commit. Abort.\n";
    return;
  }

  git_do($repo, 'branch', $branch_output);
  $branch_was = trim($branch_output[0]);
  echo "Current branch: $branch_was\n";

  git_do($repo, 'checkout master');
  git_do($repo, 'pull');
  git_do($repo, 'merge -m"Merge"');

  $tags = [];
  git_do($repo, 'tag', $tags, false);

  $last_tag = _get_last_version_tag($tags);
  echo "Last tag: $last_tag\n";

  if(empty($force_new_tag))
  {
    $last_version = GitVersion::parse($last_tag);
    $last_version->bump($up_mode);
    $new_tag = $last_version->encode();
  }
  else
    $new_tag = $force_new_tag;

  echo "New tag: $new_tag\n";

  if($branch_was != "master")
    git_do($repo, 'cherry-pick ' . $rev);

  git_do($repo, 'tag ' . $new_tag);

  git_do($repo, 'push origin master');

  git_do($repo, 'push --tags');

  git_do($repo, 'rev-parse HEAD', $output);
  if(_composer_try_update_lock_entry($relpath, trim($output[0]), $new_tag))
    return;
  
  echo "composer.lock file was not updated, run 'composer update $relpath'?\n";
  are_you_sure();

  composer_run(['update', $relpath]);
});

task('composer_vendor_status', function(array $args) {
  global $GAME_ROOT;

  $cmd = 'status -s';
  if(isset($args[0]) && $args[0] == '-p')
  {
    $cmd = 'diff';
    array_shift($args);
  }

  $project_filter = '*/*';
  if(isset($args[0]))
    $project_filter = $args[0];

  $items = glob("$GAME_ROOT/composer/vendor/$project_filter");

  foreach($items as $item)
  {
    if(is_dir($item) && is_dir("$item/.git"))
    {
      $status = [];
      exec("cd $item && git $cmd", $status);
      if($status)
      {
        $package = str_replace(normalize_path("$GAME_ROOT/composer/vendor/"), '', normalize_path($item));
        $last_remote_tag = get_git_last_remote_tag($item);
        echo "==== Vendor package '$package'[$last_remote_tag] changes";
        if(is_file($item . '/package.json'))
          echo " (UPM package?)";
        echo ":\n";
        echo implode("\n", $status) . "\n";
      }
    }
  }
  
});

function _composer_try_update_lock_entry(string $package, string $rev, string $new_tag)
{
  global $GAME_ROOT;

  $file = "$GAME_ROOT/composer/composer.lock";

  $changed = false;

  if(!is_file($file))
    return $changed;

  $json = json_decode(ensure_read($file), true);

  if(!isset($json['packages']))
    return $changed;

  foreach($json['packages'] as &$item)
  {
    if(!isset($item['source']['reference']))
      continue;

    if($item['name'] == $package)
    {
      $changed = true;
      $item['version'] = $new_tag;
      $item['source']['reference'] = $rev;

      $tz = date_default_timezone_get();
      date_default_timezone_set('UTC');
      $item['time'] = date(DATE_ATOM);
      date_default_timezone_set($tz);

      break;
    }
  }

  if($changed)
    ensure_write($file, json_encode($json, JSON_PRETTY_PRINT | JSON_UNESCAPED_SLASHES));

  return $changed;
}

function composer_run(array $args)
{
  global $GAME_ROOT;
  putenv('COMPOSER_RUN='.implode(' ', $args));
  include("$GAME_ROOT/composer/update.php");
}

//NOTE: keeping this function for BC
function composer_update($update = false)
{
  if($update)
    composer_run(['update']);
  else
    composer_run(['install']);
}

